@@ -20,6 +20,13 @@ import requests
 
 from .ssdp import evaluate_scpd_xml
 
+from .audyssey import (
+    Audyssey,
+    REF_LVL_OFFSET_MAP_LABELS,
+    MULTI_EQ_MAP_LABELS,
+    DYNAMIC_VOLUME_MAP_LABELS,
+    )
+
 _LOGGER = logging.getLogger("DenonAVR")
 
 DEVICEINFO_AVR_X_PATTERN = re.compile(
@@ -53,12 +60,10 @@ ALL_ZONE_STEREO = "ALL ZONE STEREO"
 
 SOUND_MODE_MAPPING = OrderedDict(
     [('MUSIC', ['PLII MUSIC', 'DTS NEO:6 MUSIC', 'DOLBY D +NEO:X M',
-                'DTS NEO:X MUSIC', 'DOLBY PL2 MUSIC', 'AURO-3D',
-                'AURO-2D SURROUND']),
+                'DTS NEO:X MUSIC', 'DOLBY PL2 MUSIC']),
      ('MOVIE', ['PLII MOVIE', 'PLII CINEMA', 'DTS NEO:X CINEMA',
                 'DTS NEO:6 CINEMA', 'DOLBY D +NEO:X C',
-                'PLIIX CINEMA', 'DOLBY PLII MOVIE', 'AURO-3D',
-                'AURO-2D SURROUND']),
+                'PLIIX CINEMA', 'DOLBY PLII MOVIE']),
      ('GAME', ['PLII GAME', 'DOLBY D +NEO:X G']),
      ('AUTO', ['None']),
      ('STANDARD', ['None2']),
@@ -88,7 +93,8 @@ SOUND_MODE_MAPPING = OrderedDict(
      ('DTS SURROUND', ['DTS SURROUND', 'DTS NEURAL:X', 'STANDARD(DTS)',
                        'DTS + NEURAL:X', 'MULTI CH IN', 'DTS-HD MSTR',
                        'DTS VIRTUAL:X', 'DTS-HD + NEURAL:X', 'DTS-HD',
-                       'DTS + VIRTUAL:X']),
+                       'DTS + VIRTUAL:X', 'DTS + DOLBY SURROUND']),
+     ('AURO3D', ['AURO-3D', 'AURO-2D SURROUND']),
      ('MCH STEREO', ['MULTI CH STEREO', 'MULTI_CH_STEREO', 'MCH STEREO',
                      'MULTI CH IN 7.1']),
      ('STEREO', ['STEREO']),
@@ -244,6 +250,7 @@ STATE_OFF = "off"
 STATE_PLAYING = "playing"
 STATE_PAUSED = "paused"
 
+MAIN_ZONE = "Main"
 NO_ZONES = None
 ZONE2 = {"Zone2": None}
 ZONE3 = {"Zone3": None}
@@ -274,10 +281,10 @@ class DenonAVR:
         self._host = host
         # Main zone just set for DenonAVR class
         if self.__class__.__name__ == "DenonAVR":
-            self._zone = "Main"
+            self._zone = MAIN_ZONE
         self._zones = {self._zone: self}
 
-        if self._zone == "Main":
+        if self._zone == MAIN_ZONE:
             self._urls = DENONAVR_URLS
         elif self._zone == "Zone2":
             self._urls = ZONE2_URLS
@@ -331,10 +338,12 @@ class DenonAVR:
         self._treble = None
         self._treble_level = None
 
+        self._audyssey = Audyssey(receiver=self)
+
         # Get initial setting of values
         self.update()
         # Create instances of additional zones if requested
-        if self._zone == "Main" and add_zones is not None:
+        if self._zone == MAIN_ZONE and add_zones is not None:
             self.create_zones(add_zones)
 
     def exec_appcommand_post(self, attribute_list):
@@ -385,9 +394,11 @@ class DenonAVR:
     def get_status_xml(self, command, suppress_errors=False):
         """Get status XML via HTTP and return it as XML ElementTree."""
         # Get XML structure via HTTP get
+        # Using a long read timeout because receiver could be quite slow
         res = requests.get("http://{host}:{port}{command}".format(
             host=self._host, port=self._receiver_port, command=command),
-                           timeout=self.timeout)
+                           timeout=(
+                               self.timeout, max(self.timeout, 10)))
         # Continue with XML processing only if HTTP status code = 200
         if res.status_code == 200:
             try:
@@ -409,9 +420,11 @@ class DenonAVR:
     def send_get_command(self, command):
         """Send command via HTTP get to receiver."""
         # Send commands via HTTP get
+        # Using a long read timeout because receiver could be quite slow
         res = requests.get("http://{host}:{port}{command}".format(
             host=self._host, port=self._receiver_port, command=command),
-                           timeout=self.timeout)
+                           timeout=(
+                               self.timeout, max(self.timeout, 10)))
         if res.status_code == 200:
             return True
         else:
@@ -423,9 +436,11 @@ class DenonAVR:
     def send_post_command(self, command, body):
         """Send command via HTTP post to receiver."""
         # Send commands via HTTP post
+        # Using a long read timeout because receiver could be quite slow
         res = requests.post("http://{host}:{port}{command}".format(
             host=self._host, port=self._receiver_port, command=command),
-                            data=body, timeout=self.timeout)
+                            data=body, timeout=(
+                                self.timeout, max(self.timeout, 10)))
         if res.status_code == 200:
             return res.text
         else:
@@ -510,8 +525,18 @@ class DenonAVR:
 
         if self._receiver_type == AVR_X_2016.type:
             return bool(self._update_avr_2016())
+        elif self._support_update_avr_2016:
+            if self._update_avr_2016() is not True:
+                _LOGGER.debug(
+                    "Update method (AppCommand.xml) failed for zone %s",
+                    self._zone)
+                return False
         else:
-            return self._update_avr()
+            if self._update_avr() is not True:
+                return False
+
+        # Media data are only supported on non AVR devices built before 2016
+        return self._set_media_state()
 
     def _update_avr(self):
         """
@@ -522,9 +547,8 @@ class DenonAVR:
         This method is for pre 2016 AVR(-X) devices
         """
         # Use ThreadPoolExecutor to call all URLs of this method in parallel
-        executor = ThreadPoolExecutor(max_workers=3)
-        if self._support_update_avr_2016:
-            update_avr = executor.submit(self._update_avr_2016)
+        executor = ThreadPoolExecutor(max_workers=2)
+
         status_status = executor.submit(self.get_status_xml, self._urls.status)
         status_mainzone = executor.submit(
             self.get_status_xml, self._urls.mainzone)
@@ -534,41 +558,27 @@ class DenonAVR:
                          "MasterVolume": None}
 
         # Sound mode information only available in main zone
-        if self._zone == "Main" and self._support_sound_mode:
+        if self._zone == MAIN_ZONE and self._support_sound_mode:
             relevant_tags["selectSurround"] = None
             relevant_tags["SurrMode"] = None
 
-        # if update_avr_2016 is supported try that first, that reports better
-        if self._support_update_avr_2016:
-            if update_avr.result():
-                # Success --> skip xml update
-                relevant_tags = {}
-            else:
-                _LOGGER.debug(
-                    "Primary update method (AppCommand.xml) "
-                    "failed for zone %s", self._zone)
-
         # Get status XML from Denon receiver via HTTP
-        if relevant_tags:
-            try:
-                root = status_status.result()
-            except requests.exceptions.ConnectTimeout:
-                # On connect timeout, the device is probably off
-                self._power = POWER_OFF
-                self._state = STATE_OFF
-            except (ValueError, requests.exceptions.RequestException):
-                pass
-            else:
-                # Get the tags from this XML
-                relevant_tags = self._get_status_from_xml_tags(root,
-                                                               relevant_tags)
+        try:
+            root = status_status.result()
+        except requests.exceptions.ConnectTimeout:
+            # On connect timeout, the device is probably off
+            self._power = POWER_OFF
+            self._state = STATE_OFF
+        except (ValueError, requests.exceptions.RequestException):
+            pass
+        else:
+            # Get the tags from this XML
+            relevant_tags = self._get_status_from_xml_tags(root, relevant_tags)
 
         # Second option to update variables from different source
         if relevant_tags and self._power != POWER_OFF:
             _LOGGER.debug(
-                "Secondary update method (Status.xml) failed for zone %s",
-                self._zone
-            )
+                "Update method (Status.xml) failed for zone %s", self._zone)
             try:
                 root = status_mainzone.result()
             except requests.exceptions.ConnectTimeout:
@@ -585,42 +595,13 @@ class DenonAVR:
         # Error message if still some variables are not updated yet
         if relevant_tags and self._power != POWER_OFF:
             _LOGGER.debug(
-                "Third update method (MainZone.xml) failed for zone %s",
-                self._zone)
+                "Update method (MainZone.xml) failed for zone %s", self._zone)
             _LOGGER.error("Missing status information from XML of %s for: %s",
                           self._zone, ", ".join(relevant_tags.keys()))
 
-        # Set state and media image URL based on current source
-        # and power status
-        if (self._power == POWER_ON) and (
-                self._input_func in self._playing_func_list):
-            if self._update_media_data():
-                pass
-            else:
-                _LOGGER.error(
-                    "Update of media data for source %s in %s failed",
-                    self._input_func, self._zone)
-        elif self._power == POWER_ON:
-            self._state = STATE_ON
-            self._title = None
-            self._artist = None
-            self._album = None
-            self._band = None
-            self._frequency = None
-            self._station = None
-            self._image_url = None
-        else:
-            self._state = STATE_OFF
-            self._title = None
-            self._artist = None
-            self._album = None
-            self._band = None
-            self._frequency = None
-            self._station = None
-
         # Get/update sources list if current source is not known yet
-        if (self._input_func not in self._input_func_list and
-                self._input_func is not None):
+        if (self._input_func not in self._input_func_list
+                and self._input_func is not None):
             if self._update_input_func_list():
                 _LOGGER.info("List of input functions refreshed.")
                 # If input function is still not known, log error.
@@ -643,10 +624,10 @@ class DenonAVR:
 
         Method queries device via HTTP and updates instance attributes.
         Returns "True" on success and "False" on fail.
-        This method is for AVR-X  devices built in 2016 and later.
+        This method is for AVR-X devices built in 2016 and later.
         """
         # Use ThreadPoolExecutor to call all URLs of this method in parallel
-        with ThreadPoolExecutor(max_workers=2) as executor:
+        with ThreadPoolExecutor(max_workers=3) as executor:
 
             success = True
             # Collect tags for AppCommand.xml call
@@ -659,6 +640,9 @@ class DenonAVR:
             # Update tone control
             executor.submit(self._update_tone_control)
 
+            # Update Audyssey
+            executor.submit(self._audyssey.update())
+
             try:
                 root = app_command.result()
             except requests.exceptions.ConnectTimeout:
@@ -709,12 +693,19 @@ class DenonAVR:
                     _LOGGER.error("No Source found for zone %s", self.zone)
                 success = False
             else:
+                # AirPlay and Internet Radio are not always listed in
+                # available sources
+                if inputfunc in ["AirPlay", "Internet Radio"]:
+                    if inputfunc not in self._input_func_list:
+                        self._input_func_list[inputfunc] = inputfunc
+                        self._input_func_list_rev[inputfunc] = inputfunc
                 try:
                     self._input_func = self._input_func_list_rev[inputfunc]
                 except KeyError:
-                    _LOGGER.info("No mapping for source %s found", inputfunc)
                     self._input_func = inputfunc
                     # Update sources list if current source is not known yet
+                    _LOGGER.info(
+                        "No mapping for source %s found", inputfunc)
                     if self._update_input_func_list():
                         _LOGGER.info("List of input functions refreshed.")
                         # If input function is still not known, log error.
@@ -725,8 +716,8 @@ class DenonAVR:
                                 self._input_func)
                     else:
                         _LOGGER.error((
-                            "Input function list for Denon receiver at host %s"
-                            " could not be updated."), self._host)
+                            "Input function list for Denon receiver at "
+                            "host %s could not be updated."), self._host)
             try:
                 self._sound_mode_raw = root[4][0].text.rstrip()
             except (AttributeError, IndexError):
@@ -743,7 +734,6 @@ class DenonAVR:
                     self._state = STATE_ON
                 else:
                     self._state = STATE_OFF
-
             return success
 
     def _update_input_func_list(self):
@@ -1196,8 +1186,7 @@ class DenonAVR:
                 for deleted_source in deleted_non_x_sources.items():
                     if deleted_source[1] == "DEL":
                         non_x_sources.pop(deleted_source[0], None)
-            # Invalid source "SOURCE" needs to be deleted
-            non_x_sources.pop("SOURCE", None)
+
             return non_x_sources
 
         # Following source determination of AVR-X receivers
@@ -1229,9 +1218,6 @@ class DenonAVR:
                                 "FuncName").text] = xml_source.find(
                                     "DefaultName").text
 
-            # Invalid source "SOURCE" needs to be deleted
-            receiver_sources.pop("SOURCE", None)
-
             return receiver_sources
 
     def _get_own_zone(self):
@@ -1241,11 +1227,49 @@ class DenonAVR:
         These zone information are used to evaluate responses of HTTP POST
         commands.
         """
-        if self._zone == "Main":
+        if self._zone == MAIN_ZONE:
             return "zone1"
         else:
             return self.zone.lower()
 
+    def _set_media_state(self):
+        """
+        Set media state for the receiver.
+
+        This method does not work on AVR-X devices built in 2016 and later.
+        """
+        # Set state and media image URL based on current source
+        # and power status
+        if (self._power == POWER_ON and
+                self._input_func in self._playing_func_list):
+            if self._update_media_data():
+                return True
+            else:
+                _LOGGER.error(
+                    "Update of media data for source %s in %s failed",
+                    self._input_func, self._zone)
+                return False
+        elif self._power == POWER_ON:
+            self._state = STATE_ON
+            self._title = None
+            self._artist = None
+            self._album = None
+            self._band = None
+            self._frequency = None
+            self._station = None
+            self._image_url = None
+        else:
+            self._state = STATE_OFF
+            self._title = None
+            self._artist = None
+            self._album = None
+            self._band = None
+            self._frequency = None
+            self._station = None
+            self._image_url = None
+
+        return True
+
     def _update_media_data(self):
         """
         Update media data for playing devices.
@@ -1286,7 +1310,7 @@ class DenonAVR:
                     self._frequency = None
                     self._station = None
 
-        elif self._input_func == "Tuner" or self._input_func == "TUNER":
+        elif self._input_func in ["Tuner", "TUNER"]:
             try:
                 root = self.get_status_xml(self._urls.tunerstatus)
             except (ValueError, requests.exceptions.RequestException):
@@ -1312,7 +1336,7 @@ class DenonAVR:
                 STATIC_ALBUM_URL.format(
                     host=self._host, port=self._receiver_port))
 
-        elif self._input_func == "HD Radio" or self._input_func == "HDRADIO":
+        elif self._input_func in ["HD Radio", "HDRADIO"]:
             try:
                 root = self.get_status_xml(self._urls.hdtunerstatus)
             except (ValueError, requests.exceptions.RequestException):
@@ -1402,6 +1426,12 @@ class DenonAVR:
             elif child.tag == "InputFuncSelect":
                 inputfunc = child[0].text
                 if inputfunc is not None:
+                    # AirPlay and Internet Radio are not always listed in
+                    # available sources
+                    if inputfunc in ["AirPlay", "Internet Radio"]:
+                        if inputfunc not in self._input_func_list:
+                            self._input_func_list[inputfunc] = inputfunc
+                            self._input_func_list_rev[inputfunc] = inputfunc
                     try:
                         self._input_func = self._input_func_list_rev[inputfunc]
                     except KeyError:
@@ -1635,6 +1665,71 @@ class DenonAVR:
         """Return level of treble."""
         return self._treble_level
 
+    @property
+    def dynamic_eq(self):
+        """Return value of Dynamic EQ."""
+        return self._audyssey.dynamiceq
+
+    @property
+    def reference_level_offset(self):
+        """Return value of Reference Level Offset."""
+        return self._audyssey.reflevoffset
+
+    @property
+    def reference_level_offset_setting_list(self):
+        """Return a list of available reference level offset settings."""
+        return list(REF_LVL_OFFSET_MAP_LABELS.keys())
+
+    @property
+    def dynamic_volume(self):
+        """Return value of Dynamic Volume."""
+        return self._audyssey.dynamicvol
+
+    @property
+    def dynamic_volume_setting_list(self):
+        """Return a list of available Dynamic Volume settings."""
+        return list(DYNAMIC_VOLUME_MAP_LABELS.keys())
+
+    @property
+    def multi_eq(self):
+        """Return value of MultiEQ."""
+        return self._audyssey.multeq
+
+    @property
+    def multi_eq_setting_list(self):
+        """Return a list of available MultiEQ settings."""
+        return list(MULTI_EQ_MAP_LABELS.keys())
+
+    def dynamic_eq_off(self):
+        """Turn DynamicEQ off."""
+        self._audyssey.dynamiceq_off()
+
+    def dynamic_eq_on(self):
+        """Turn DynamicEQ on."""
+        self._audyssey.dynamiceq_on()
+
+    def toggle_dynamic_eq(self):
+        """Toggle DynamicEQ."""
+        if self.dynamic_eq is True:
+            self._audyssey.dynamiceq_off()
+        else:
+            self._audyssey.dynamiceq_on()
+
+    @reference_level_offset.setter
+    def reference_level_offset(self, setting):
+        """Set Reference Level Offset."""
+        self._audyssey.set_reflevoffset(setting=setting)
+
+    @dynamic_volume.setter
+    def dynamic_volume(self, setting):
+        """Set Dynamic Volume."""
+        self._audyssey.set_dynamicvol(setting=setting)
+
+    @multi_eq.setter
+    def multi_eq(self, setting):
+        """Set MultiEQ."""
+        self._audyssey.set_mutlieq(setting=setting)
+
     @input_func.setter
     def input_func(self, input_func):
         """Setter function for input_func to switch input_func of device."""
@@ -1924,6 +2019,11 @@ class DenonAVR:
             _LOGGER.error("Connection error: set volume command not sent.")
             return False
 
+    @volume.setter
+    def volume(self, volume):
+        """Set receiver volume. Minimum is -80.0, maximum at 18.0."""
+        self.set_volume(volume)
+
     def mute(self, mute):
         """Mute receiver via HTTP get command."""
         try:
@@ -1957,13 +2057,13 @@ class DenonAVR:
             self.send_post_command(
                 command=self._urls.appcommand, body=body.getvalue()
                 )
-        except (
-                requests.exceptions.ConnectTimeout,
-                requests.exceptions.RequestException
-                ):
+        except requests.exceptions.RequestException:
             _LOGGER.error("No connection to %s end point on host %s",
                           self._urls.appcommand, self._host)
             return False
+        finally:
+            # Buffered XML not needed anymore: close
+            body.close()
         return True
 
     def enable_tone_control(self):
@@ -2110,3 +2210,56 @@ class DenonAVRZones(DenonAVR):
         Return "True" on success and "False" on fail.
         """
         return self._parent_avr.set_sound_mode(sound_mode)
+
+    def _get_renamed_deleted_sources(self):
+        """
+        Get renamed and deleted sources lists from receiver .
+
+        Internal method which queries device via HTTP to get names of renamed
+        input sources.
+        """
+        renamed, deleted = super()._get_renamed_deleted_sources()
+
+        # SOURCE in zone 2 and 3 means "same input sources as main zone"
+        renamed["SOURCE"] = "Source of main zone"
+
+        return (renamed, deleted)
+
+    def _get_renamed_deleted_sourcesapp(self):
+        """
+        Get renamed and deleted sources lists from receiver .
+
+        Internal method which queries device via HTTP to get names of renamed
+        input sources. In this method AppCommand.xml is used.
+        """
+        renamed, deleted, success = super()._get_renamed_deleted_sourcesapp()
+
+        if success is True:
+            # SOURCE in zone 2 and 3 means "same input sources as main zone"
+            renamed["SOURCE"] = "Source of main zone"
+
+        return (renamed, deleted, success)
+
+    def _set_media_state(self):
+        """
+        Set media state for the receiver.
+
+        This method does not work on AVR-X devices built in 2016 and later.
+        """
+        if (self._parent_avr.input_func in self._parent_avr.playing_func_list
+                and self._input_func == "Source of main zone"
+                and self._power == POWER_ON):
+
+            # When zone uses the same input as main zone, copy its media state
+            self._state = self._parent_avr.state
+            self._title = self._parent_avr.title
+            self._artist = self._parent_avr.artist
+            self._album = self._parent_avr.album
+            self._band = self._parent_avr.band
+            self._frequency = self._parent_avr.frequency
+            self._station = self._parent_avr.station
+            self._image_url = self._parent_avr.image_url
+
+            return True
+        else:
+            return super()._set_media_state()
